package android.lib.database;

import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Marks a field as a table column.
 * <p>The class containing the annotated field must be annotated by {@link Table @Table}.</p>
 * <p>An annotated field can also be annotated by {@link Index @Index}.</p>
 * <p><p>SQLite-compatible types include <code>boolean</code>, <code>byte</code>, <code>byte[]</code>,
 * <code>double</code>, <code>float</code>, <code>int</code>, <code>long</code>, <code>short</code>
 * and <code>String</code>. If the type of the annotated field is not SQLite-compatible,
 * a {@link TypeConverter} is required for type conversion. Use <{@link UseConverter @UseConverter}
 * to indicate the {@link TypeConverter} to use for the annotated field.</p>
 * @see Table
 * @see Index
 * @see UseConverter
 */
@Target({ ElementType.FIELD })
@Retention(RetentionPolicy.RUNTIME)
public @interface Column {
    /**
     * The name of the table name.
     * <p>If not specified, the name in exact case of the annotated field will be used.</p>
     */
    String value() default "";

    /**
     * <code>true</code> if the column can be <code>null</code>; otherwise, <code>false</code>.
     * <p>{@link nullable} is ignored if {@link primaryKey} is <code>true</code>.</p>
     */
    boolean nullable() default true;

    /**
     * <code>true</code> if the column is a primary key; otherwise, <code>false</code>.
     * <p>Only one column within a table can be defined as primary key.</p>
     */
    boolean primaryKey() default false;

    /**
     * <code>true</code> if the primary key column is also an auto-increment column,
     * otherwise, <code>false</code>.
     * <p>{@link autoIncrement} is ignored if {@link primaryKey} is not set to <code>true</code>.</p>
     */
    boolean autoIncrement() default false;
}
